/******************************************************************
 * Copyright (c) 1990  MOTOROLA, Inc.                             *
 * NOTICE: The information contained in this file is proprietary  *
 * to MOTOROLA, INC. and is being made available  to MOTOROLA'S   *
 * customers under strict license agreements.              *
 * Use or disclosure of this information is permissible only      *
 * under the terms of the existing license agreement.      *
 ******************************************************************/

/* LOCALTALK ADAPTOR HEADER FILE
/*	LLOYD HASLEY   7/2/90
/*Copyright 1990  Motorola SPS
/*
/* based on Appendix B:LLAP Access Control Algorithm
	of Inside AppleTalk, Second Edition              */
/* */



/**** USERS OF THE LOCAL-TALK SAMPLE DRIVER NEED TO SET
	THE FOLLOWING VARIABLES *********************************/
/* set main pointers */
#define	basep  ((IMP *) 0x700000) /* base address of 68302 */
#define	PORT0			/* set PORT0 or PORT1 */
/* Options: */
#define	STATISTIC		/* enable STATISTIC gathering */
#define	TIMING			/* if defined, puts out timing marks, */
				/* during transmit and receive interrupt */

/***************END OF USER SETTABLE PORTION ************************/


/* global LA definitions */
#define minFrameSize 	3	/* smallest (LAP header only) frame */
#define	maxFrameSize 	605	/* size of largest LAP frame including FCS*/
#define	maxDataSize 	600	/* size of largest (encapsulated) LLAP data field */
#define	bitTime		4.34	/* bit time (uS) */
#define	BitTimeNs	4340	/* bit time (nS) */
#define	byteTime	39	/* worst case single byte time (uS) */
#define minIDGtime      400   /* minimum interdialog gap (uS) */
#define	IDGslottime	100	/* slot time of transmit backoff algorithm (uS) */
#define	maxIFGtime	200	/* maximum interfame gap (uS) */
#define	minIFGtime	50	/* minimum interface gap (uS) */
#define	maxDefers	32	/* maximum defers for a single packet */
#define	maxCollsns	32	/* maximum collisions for a single packet */

#define	wksTries	20	/* # of ENQ sets for a workstation to try */

#define	lapENQ		0x81	/* LAP type field: ENQuiry frame */
#define lapACK          0x82    /* LAP type field: ACKnowledgment frame */
#define lapRTS          0x84    /* LAP type field: RequestToSend frame */
#define lapCTS          0x85    /* LAP type field: ClearToSend frame */

#define	lapDATA		0xa0	/*LAP Data frame has been sent by recv intr*/

#define	hdlcFLAG	0x7e	/* value of a HDLC flag */
#define	BROADCAST	0xff	/* network broadcast address */


#define	MAXRBUF		20		/* number of receive frames held in queue */


/* result codes from LAP functions */
/* TransmitStatus */
#define	notInitialize	0
#define	transmitOK	1
#define	receiveOK	1
#define	excessDefers	2
#define	excessCollsns	3
#define	dupAddress	4
#define	networkError	5
#define	OKinitialize	6

/* lapresult is string equivalents of above.  any change in above definitions
	then lapresult initialization must also be editied */
extern char lapresult[][20];

/* data link types and structures */
typedef UCHAR	octet;
typedef octet	anAddress;
typedef	octet	aLAPtype;
typedef	octet	aDataField[maxDataSize];
typedef	octet	aRDataField[maxFrameSize];

/* basic LLAP frame structure, w/o flags, fcs */
typedef union {
	octet	raw[maxFrameSize];
	struct cooked {
		anAddress dstAddr;
		anAddress srcAddr;
		aLAPtype  lapType;
		aDataField dataField;
	} structured ;
} aFrame;
/* frame structure, same as above but with fcs */
typedef union {
	octet	raw[maxFrameSize];
	struct rcooked {
		anAddress dstAddr;
		anAddress srcAddr;
		aLAPtype  lapType;
		aRDataField dataField;
	} structured ;
} aRFrame;


/* transmit uses only two descriptors.  BD0 is used for packet headers
	BD1 is used for packet payloads (if any) */
typedef union {			/* transmit BD 0 template */
	struct headerc {
		anAddress dstAddr;
		anAddress srcAddr;
		aLAPtype  lapType;
	} structured;
	UINT32	raw;	/*must be bigger than structured structure*/
} theader;


/* receive uses a link list buffer structure between ReceivePacket and
	the actual interrupt handler */
struct rbuf {
	struct rbuf *next;				/* pointer to next item in list */
	UINT16 length;				/* length of receive packet */
	aRFrame rframe;			/* actual receive frame */
} ;

/* xmit buffer descriptor utilization */
#define	XBD_NULL	0	/* reserved, always empty */
#define	XBD_ENQ		1	/* pointer to a generic ENQ frame */
#define	XBD_ACK		2	/* pointer to a generic ACK frame */
#define	XBD_RTS		3	/* pointer to a generic RTS frame */
#define	XBD_CTS		4	/* pointer to a generic CTS frame */
#define	XBD_HEAD	5	/* pointer to data frame header */
#define	XBD_DATA	6	/* pointer to actual data field */

/* LLAP global variables */
extern	octet	MyAddress;		/* set during InitializeLAP */
extern	int	Backoff;		/* current backoff range */
extern	Boolean	fAdrValid;		/* MyAddress has been validated */
extern	Boolean	fAdrInUse;		/* Another node has same MyAddress */
extern	Boolean	fCTSexpected;		/* RTS has been sent, CTS is valid */
extern struct hdlc_pram *hdlcp;
extern	IMP_BD *rbd_p;			/* receive buffer[0] pointer */
#define	R_STAT_DEFAULT	(R_X|R_W)	/* default rbd contents */
extern	int RX_Frame_Status;	/* semaphore to transmit routine */


#ifdef	STATISTIC
extern	UINT16 deferCount, collsnCount;	/* optional, for statistics only */
extern UINT32 RFrameCount,RDataCount; /* receive frame statisics */
extern UINT32 ROvrFlow;		/* no more buffer counter */
#endif
extern	UCHAR	deferHistory;
extern	UCHAR	collsnHistory;
extern	int outgoingLength;
extern	aFrame outgoingPkt;
extern int debug;
extern int sdebug;
extern int init_complete;
extern struct rbuf *RPacket();

#ifdef	PORT1
#define	DATADEF		0x1073		/* debugging stuff */
#define	DATAMARK	0x1273		/* allows toggling of i/o */
#endif
#ifdef	PORT0
#define	DATADEF		0x1073		/* debugging stuff */
#define	DATAMARK	0x1473		/* allows toggling of i/o */
#endif

/* port dependent stuff, note: this is configured from the user
	portion at the top of this file */
#ifdef	PORT0
#define	PORT	0
#else
#define	PORT	1
#endif


/* MISC Macro Definitions */
#define	RANDOM(maxrange)	(rand()%maxrange)
#define	MIN(val1,val2)		( (val1<val2)?val1:val2 )
#define	MAX(val1,val2)		( (val1>val2)?val1:val2 )
#define RstCarrierSense()	(basep->scc_regs[PORT].event = HDLC_CD);
