/******************************************************************
 * Copyright (c) 1990  MOTOROLA, Inc.                             *
 * NOTICE: The information contained in this file is proprietary  *
 * to MOTOROLA, INC. and is being made available  to MOTOROLA'S   *
 * customers under strict license agreements.              *
 * Use or disclosure of this information is permissible only      *
 * under the terms of the existing license agreement.      *
 ******************************************************************/

#include <stdio.h>
#include "system.h"				/* variable type defines */
#include "imp.h"
#include "hdlc.h"
#include "la.h"


/******************************************************************
 *                                                                *
 *  This file is in four parts:                                   *
 *      1)  Global Variable Declarations                          *
 *      2)  Receive Routines                                      *
 *      3)  Hardware Initialization/Interface (low-level)         *
 *      4)  Transmit Routines                                     *
 *                                                                *
 *                                                                *
 *                                                                *
 *  ************************************************************  *
 *     Part 1:  Global Variable Declarations                      *
 *                                                                *
 ******************************************************************/

struct hdlc_pram *hdlcp;
IMP_BD *rbd_p;			/* receive buffer desc. pointer */
int init_complete = FALSE;
int old_intr;

/* LLAP global variables */
octet	MyAddress;		/* set during InitializeLAP */
int	Backoff;		/* current backoff range */
Boolean	fAdrValid;	/* MyAddress has been validated */
Boolean	fAdrInUse;	/* Another node has same MyAddress */
Boolean	fCTSexpected;		/* RTS has been sent, CTS is valid */

#ifdef	STATISTIC
UINT16 deferCount, collsnCount;	/* optional, for statistics only */
#endif

UCHAR	deferHistory;
UCHAR	collsnHistory;

char lapresult[][20] = { "Not Initialized","Transmit OK","Excess Deferals",
	"Excess Collisions","Duplicate Address","Network Error" } ;

struct rbuf rbuffer[MAXRBUF];	/* allocate receive buffers */
struct rbuf freelist;		/* header for free list */
struct rbuf rhdr;		/* header for receive list */
struct rbuf rtail;		/* tail pointer for receive list */

int RX_Frame_Status=0;		/* semaphore to transmit routine */
				/* 0 - still waiting for response */
				/* lapACK - ack received */
				/* lapCTS - cts received */

#ifdef	STATISTIC
UINT32 RFrameCount,RDataCount;	/* receive frame statisics */
UINT32 ROvrFlow;		/* no more buffer counter */
#endif



/******************************************************************
 *                                                                *
 *  This file is in four parts:                                   *
 *      1)  Global Variable Declarations                          *
 *      2)  Receive Routines                                      *
 *      3)  Hardware Initialization/Interface (low-level)         *
 *      4)  Transmit Routines                                     *
 *                                                                *
 *                                                                *
 *                                                                *
 *  ************************************************************  *
 *     Part 2:  Receive Routines                                  *
 *                                                                *
 ******************************************************************/


/***************************************************************************/
/*  Procedure: la_recv_init
	initialize the recv link lists.
	note: the receive buffer space is statically allocated.
	this allows re-init without extra memory being allocated.

	for most applications this routine will need to be modified to
	support system memory allocation, init, and re-init
*/
/***************************************************************************/
la_recv_init()
{
	struct rbuf *p;

	for(p= (struct rbuf *)rbuffer;p< &(rbuffer[MAXRBUF-1]); p++)
		p->next = p + 1;
	p->next = (struct rbuf *)0;

	freelist.next = (struct rbuf *)rbuffer;
	rhdr.next = (struct rbuf *)0;
	rtail.next = (struct rbuf *)0;

	/* MISC RECV INITS */
	RX_Frame_Status=0;		/* semaphore to transmit routine */
#ifdef	STATISTIC
	RFrameCount=RDataCount=0L;	/* receive frame statisics */
	ROvrFlow=0L;
#endif
}

/***************************************************************************/
/*  Procedure: ReceivePacket
	Primary Interface routine to higher levels.
*/
/***************************************************************************/
ReceivePacket(dstParam,srcParam,typeParam,dataParam,dataLength)
anAddress *dstParam;
anAddress *srcParam;
aLAPtype *typeParam;
octet *dataParam;
int *dataLength;
{
	int i;
	aRFrame *p;
	struct rbuf *rp;
	extern struct rbuf *RPacket();

	if(!init_complete) {
		if(sdebug>0)
			printf("la driver not initialized!!\n");
		return(notInitialize);
	}
	while((rp=RPacket()) == (struct rbuf *)0);  /*wait for receive packet*/

	p = &(rp->rframe);

	*dstParam = p->structured.dstAddr;
	*srcParam = p->structured.srcAddr;
	*typeParam = p->structured.lapType;
	for(i=0;i<rp->length;i++)
		*dataParam++ = p->structured.dataField[i];
	*dataLength = rp->length;
	FreePacket(rp);   /* done with packet buffer, put back on free list */

	return(receiveOK);
}

/***************************************************************************/
/*  Procedure: RPacket
		Alternate Receive Interface routine to higher levels.
		Unlike ReceivePacket, which copies the frame, RPacket returns
			a pointer to the rbuf structure.
	Note:  If RPacket is used, instead of ReceivePacket, then FreePacket
		must be called, eventually to return the buffer the receive routines
*/
/***************************************************************************/
struct rbuf *
RPacket()
{
	register struct rbuf *rp;

	/* remove entry from receive list */
	rp = rhdr.next;
	if(rp == (struct rbuf *)0)
		return( (struct rbuf *)0);
	la_di();			/* disable interrupts */
 	rhdr.next = rp->next;
	if(rhdr.next == (struct rbuf *)0)	/* only one item on list */
		rtail.next = (struct rbuf *)0;	/* clear tail pointer */
	la_ei();								/* re-enable interrupts */
	
	/* take out addresses,type,crc from length */
	/* should be in interrupt, but here reduces interrupt time */
	rp->length = ((INT16)(rp->length)) - 5;
	if( ((INT16)(rp->length)) <0)
		rp->length = 0;

	return(rp);
}

/***************************************************************************/
/*  Procedure: RPacketAvail
	Returns non-zero if there is a packet on the receive list
	Retruns zero if is NO packet on the receive list
*/
/***************************************************************************/
RPacketAvail()
{
	if(rhdr.next== (struct rbuf *)0)
		return(FALSE);
	  else
		return(TRUE);
}

/***************************************************************************/
/*  Procedure: FreePacket
	Companion to the Alternate Receive Interface routine, RPacket.
	FreePacket returns the buffer the receive routines
*/
/***************************************************************************/
FreePacket(rp)
struct rbuf *rp;
{
	/* put structure onto free list */
	la_di();								/* disable interrupts */
	rp->next = freelist.next;
	freelist.next = rp;
	la_ei();								/* re-enable interrupts */
}

/***************************************************************************/
/*  Procedure: GetRBuf
	takes a buffer, if any, from the free list and returns pointer
*/
/***************************************************************************/
struct rbuf *
GetRBuf()
{
	struct rbuf *p;

	if(freelist.next == (struct rbuf *)0)
		return( (struct rbuf *)0);

	la_di();
	p = freelist.next;
	freelist.next = p->next;
	la_ei();

	return(p);
}

/***************************************************************************/
/*  Procedure: ReceiveFrame
	receive interrupt handler
*/
/***************************************************************************/
_IH void
Intr_ReceiveFrame()
{
	register aRFrame *p;
	register struct hdlc_mode_reg *hdlc_mp;
	struct rbuf *rbufp;

#ifdef	TIMING
basep->pa_data=DATAMARK;
#endif
	/* need to sync transmit polls, so cycle txen bit */
	hdlc_mp = (struct hdlc_mode_reg *) &(basep->scc_regs[PORT].mode);
	hdlc_mp->tx_enable = 0;
	hdlc_mp->flg = 1;		/* set scc poll every 8 tclks */

	/* event register: RXF,BSY, or RXB */
	/* i.e. we are here only if we have a receive interrupt */
	/* so clear interrupt */
	basep->scc_regs[PORT].event = HDLC_RXFR|HDLC_BUSY|HDLC_RXBD;
	hdlc_mp->tx_enable = 1;	/* give time for cp to recognize off state
					before turning it back on */
#ifdef	PORT0
	basep->intr_inservice = 0x2000;
#else	/* PORT1 */
	basep->intr_inservice = 0x0400;
#endif

	/* check receive buffer, is it empty */
	if(basep->pram[PORT].scc.rbd[0].status & R_E)
		return;		/* false intr, buffer still owned by risc*/
	/* do we have an error ? */
	if(basep->pram[PORT].scc.rbd[0].status & R_ERROR) {/* have an error */
retbuf:
		/* ignore frame, return buffer to risc */
		rbd_p->status = R_STAT_DEFAULT|R_E;
		return;
	}
	if(basep->pram[PORT].scc.rbd[0].buf[1] == (octet)0)/* null address */
		goto retbuf;
	if(basep->pram[PORT].scc.rbd[0].length<minFrameSize)	/* length error */
		goto retbuf;		/* note: max length checked by mflr */
	p = (aRFrame *) rbd_p->buf;

	if(fAdrValid)  {	/* if our address is valid, check on frame type */
#ifdef	STATISTIC
		RFrameCount++;		/* receive frame statisics */
#endif
		if(p->structured.lapType < 0x80) {			/* it is a data frame */
#ifdef	STATISTIC
			RDataCount++;	/* receive data packet statisics */
#endif
			if(freelist.next != (struct rbuf *)0) {
				/* if we are going to have other buffer */
				/* take this frame, if no more buffers then */
				/* don't give up buffer, drop current frame
					onto the proverbial floor */

				/* put frame onto received link list */
				rbufp = (struct rbuf *)
				   (((char *)p)-sizeof(struct rbuf *)-sizeof(UINT16));
				rbufp->next = (struct rbuf *)0;
				if(rtail.next == (struct rbuf *)0) /* first */
					rhdr.next = rbufp;
				  else
					rtail.next->next = rbufp;
				rtail.next = rbufp;
				rbufp->length = rbd_p->length;
				/*done processing old message, put new buffer into descriptor*/
				rbufp = freelist.next;
				freelist.next = rbufp->next;
				rbd_p->buf = ((unsigned char *) &(rbufp->rframe));
			} else
				ROvrFlow++;	/* no more buffers */
		} else	{			/* it is a control frame */

#ifdef	TIMING
basep->pa_data=DATADEF;
#endif
			switch(p->structured.lapType) {
			case lapCTS:	   /* cts, either to our rts or oops */
				if(fCTSexpected || fAdrInUse) {
					/* if we are not trying a broadcast or
					a enq, send the frame */
					if(basep->pram[PORT].scc.tbd[XBD_HEAD].buf[0] == BROADCAST)
						RX_Frame_Status = lapCTS;
					else
					if(basep->pram[PORT].scc.tbd[XBD_HEAD].buf[2] == lapENQ)
						RX_Frame_Status = lapCTS;
					else {
						RTransmitFrame(XBD_HEAD);
						RX_Frame_Status = lapDATA;
					}
				} else	{
					fAdrInUse = TRUE;
				}
				break;
			case lapRTS:	    /* someone wants to send to us */
				/* ADDITION TO APPLE's INSIDE APPLETALK CODE*/
				/* original version sends CTS on broadcast */
				/* this is in error */
				if(p->structured.dstAddr == 0xff)
					break;
				/* END OF ADDITION */
				if(fAdrValid)
					RTransmitFrame(XBD_CTS);
				  else
					fAdrInUse = TRUE;
				break;
			case lapENQ:	   /* need to send acknowledgement */
				if(fAdrValid)
					RTransmitFrame(XBD_ACK);
				  else
					fAdrInUse = TRUE;
				break;
			case lapACK:	    /* ack, probably due to our ENQ */
				RX_Frame_Status = lapACK;
				fAdrInUse = TRUE;
				break;
			}
		}
	} else	if(p->raw[0] != BROADCAST)
		fAdrInUse = TRUE;	/* we received something we didn't expect */

	/* return buffer to risc */
	rbd_p->status = R_STAT_DEFAULT | R_E;
	hdlc_mp->flg = 0;		/* set scc poll every 16 tclks */

#ifdef	TIMING
basep->pa_data=DATAMARK;
basep->pa_data=DATADEF;
basep->pa_data=DATAMARK;
basep->pa_data=DATADEF;
#endif

}

/***************************************************************************/
/*  Procedure: RTransmitFrame
		response transmit, called due to rts or enq received
		calls transmitframe.
*/
/***************************************************************************/
RTransmitFrame(FrameType)
int FrameType;
{
	struct hdlc_mode_reg *hdlc_mp;
	int oldtstate;

	/* is transmitter currently active? */
	/* wait for risc to finish current transmit frame*/
	while(basep->pram[PORT].scc.pscc.h.tbuf != 0x40);

	/* configure send frame */
	basep->pram[PORT].scc.tbd[FrameType].buf[0] = rbd_p->buf[1];

	/* disable Receiver */
	hdlc_mp = (struct hdlc_mode_reg *) &(basep->scc_regs[PORT].mode);
	hdlc_mp->rx_enable = 0;
	while(basep->cr & 1);		/* wait for command idle */
	basep->scc_regs[PORT].event = HDLC_RXFR|HDLC_BUSY|HDLC_RXBD;
	
	oldtstate = basep->pram[PORT].scc.pscc.h.tstate;
	if(FrameType==XBD_HEAD)	{ 	/* data packet */
		if(basep->pram[PORT].scc.tbd[XBD_DATA].length==0) {
			basep->pram[PORT].scc.tbd[XBD_HEAD].status |= (T_W|T_L);
		} else	{
			basep->pram[PORT].scc.tbd[XBD_DATA].status |= T_R;
		}
	}
	basep->pram[PORT].scc.tbd[FrameType].status |= T_R; /* send frame */
	basep->pram[PORT].scc.pscc.h.tbuf = 0x40 + (FrameType<<3);

#ifdef	TIMING
basep->pa_data=DATAMARK;
#endif

	hdlc_mp->flg = 1;
	while(oldtstate==basep->pram[PORT].scc.pscc.h.tstate) ;
	hdlc_mp->flg = 0;
	basep->cr = ENTER_HUNT_MODE | (PORT<<1) | 1;

	/* wait for transmission */
	if(FrameType==XBD_HEAD)	{ 	/* data packet */
		if(basep->pram[PORT].scc.tbd[XBD_DATA].length==0) {
			while(basep->pram[PORT].scc.tbd[XBD_HEAD].status & T_R);
		} else
			while(basep->pram[PORT].scc.tbd[XBD_DATA].status & T_R);
	} else
		while(basep->pram[PORT].scc.tbd[FrameType].status & T_R);

	/* 302 guanrantees frame completely transmitted before clearing
		buffer descriptor.  This concerns the data field only.
		closing flag and the idle sequence may still be being sent.*/

	/* enable Receiver */
	while(basep->cr & 1); /* wait for hunt completion */
	hdlc_mp->rx_enable = 1;

	/* wait for frame trailer IDLE-seq */
		/* incls la 1.5 squelch timer and la internal delays */
	for(STimer(88);!GTimer(););
	basep->pram[PORT].scc.tbd[XBD_HEAD].status &= ~(T_W|T_L);

#ifdef	TIMING
basep->pa_data=DATADEF;
#endif

}

/******************************************************************
 *                                                                *
 *  This file is in four parts:                                   *
 *      1)  Global Variable Declarations                          *
 *      2)  Receive Routines                                      *
 *      3)  Hardware Initialization/Interface (low-level)         *
 *      4)  Transmit Routines                                     *
 *                                                                *
 *                                                                *
 *                                                                *
 *  ************************************************************  *
 *     Part 3:  Hardware Initialization/Interface                 *
 *                                                                *
 ******************************************************************/


/***************************************************************************/
/*  Procedure: mc302_init
	set hardware registers to desired configuration
		baseaddr  is starting location of mc68302
*/
/***************************************************************************/
mc302_init()
{
	struct scc_conf *confp;
	struct hdlc_mode_reg *hdlc_mp;
	UINT32 *udatap;
	struct rbuf *rbufp;
	int vector;
	int orig_intr_priority;		/* original interrupt priority level */
#ifdef	GNU
	extern	_INTR_SR();
#endif
	
	init_complete = FALSE;
	orig_intr_priority = _GPL(); /* original interrupt priority level */
	_SPL(7);		/* turn off all interrupts */

	hdlcp = (struct hdlc_pram *) &(basep->pram[PORT].scc.pscc.h);

	/* parallel I/O Control, assumes Motorola WW Breadboard */
#ifdef	PORT0
	basep->pa_data = 0x1073;
	basep->pa_control = 0x007f;
	basep->pa_direction = 0x1f80;
#else	/* PORT1 */
	basep->pa_data = 0x1873;
	basep->pa_control = 0x007f;
	basep->pa_direction = 0x1f80;
#endif

	/* timer 0 initialization,  1uS per tick; no interrupts */
	basep->t0_mode = 0x000d;	/* restart and V16 */
	basep->t0_ref = 0xffff;		/* max timeout */

	/* 302 configuration */
	basep->si_mode  = 0;
 
	/* SCON Register */
	confp = (struct scc_conf *) &(basep->scc_regs[PORT].conf);
	confp->woms = 0;
	confp->txclk = 1;
	confp->rxclk = 1;

	/* SCC Mode Register */
	basep->scc_regs[PORT].mode = 0x1000;
	hdlc_mp = (struct hdlc_mode_reg *) &(basep->scc_regs[PORT].mode);

	/* CREATE TRANSMIT HEADER DATA BUFFER and transmit descriptors */
	udatap = (UINT32 *)  &(basep->udata[0]);
	basep->pram[PORT].scc.tbd[XBD_NULL].status = T_W|T_L|T_CRC;
	basep->pram[PORT].scc.tbd[XBD_NULL].length = 0;
	basep->pram[PORT].scc.tbd[XBD_NULL].buf = (UCHAR *)0;
	basep->pram[PORT].scc.tbd[XBD_ENQ].status = T_W|T_L|T_CRC;
	basep->pram[PORT].scc.tbd[XBD_ENQ].length = 3;
	basep->pram[PORT].scc.tbd[XBD_ENQ].buf = (UCHAR *)udatap;
	*udatap++ = lapENQ<<8;
	basep->pram[PORT].scc.tbd[XBD_ACK].status = T_W|T_L|T_CRC;
	basep->pram[PORT].scc.tbd[XBD_ACK].length = 3;
	basep->pram[PORT].scc.tbd[XBD_ACK].buf = (UCHAR *)udatap;
	*udatap++ = lapACK<<8;
	basep->pram[PORT].scc.tbd[XBD_RTS].status = T_W|T_L|T_CRC;
	basep->pram[PORT].scc.tbd[XBD_RTS].length = 3;
	basep->pram[PORT].scc.tbd[XBD_RTS].buf = (UCHAR *)udatap;
	*udatap++ = lapRTS<<8;
	basep->pram[PORT].scc.tbd[XBD_CTS].status = T_W|T_L|T_CRC;
	basep->pram[PORT].scc.tbd[XBD_CTS].length = 3;
	basep->pram[PORT].scc.tbd[XBD_CTS].buf = (UCHAR *)udatap;
	*udatap++ = lapCTS<<8;
	basep->pram[PORT].scc.tbd[XBD_HEAD].status = T_CRC;
	basep->pram[PORT].scc.tbd[XBD_HEAD].length = 3;
	basep->pram[PORT].scc.tbd[XBD_HEAD].buf = (UCHAR *)udatap;
	*udatap++ = 0L;
	basep->pram[PORT].scc.tbd[XBD_DATA].status = T_X|T_W|T_L|T_CRC;
	basep->pram[PORT].scc.tbd[XBD_DATA].length = 0;
	basep->pram[PORT].scc.tbd[XBD_DATA].buf = (UCHAR *)0;

	/* receive buffer init */
	rbd_p = (IMP_BD *) &(basep->pram[PORT].scc.rbd[0]);
	rbufp = GetRBuf();
	rbd_p->buf = ((unsigned char *) &(rbufp->rframe));
	rbd_p->status = R_STAT_DEFAULT|R_E;
	
	/* clear SCC event/mask registers */
	basep->scc_regs[PORT].event = 0xff;	/* write one to clear */
	basep->scc_regs[PORT].mask = 0;		/* no interrupts */
	basep->scc_regs[PORT].sync = 0x7e7e;	/* SCC sync register */

	hdlcp->rfc = 0x50;			/* Rx FC register */
	hdlcp->tfc = 0x50;			/* Tx FC register */
	hdlcp->rblen = maxFrameSize;		/* Rx buffer length */

	hdlcp->r_mask = 0xff;
	hdlcp->r_addr1 = BROADCAST;		/* tell 302 recv CP process */
	hdlcp->r_addr2 = BROADCAST;
	hdlcp->r_addr3 = BROADCAST;
	hdlcp->r_addr4 = BROADCAST;
	SetAddress(BROADCAST);		/* put in transmit descriptors*/

	hdlcp->r_cmaskl = 0xf0b8;
	hdlcp->r_cmaskh = 0xf0b8;
	hdlcp->discard_cnt = 0;
	hdlcp->crc_cnt = 0;
	hdlcp->abort_cnt = 0;
	hdlcp->nmf_cnt = 0;
	hdlcp->retx_cnt = 0;
	hdlcp->r_max = maxFrameSize;

	/* activate receive interrupt */
	/* select user vectors*/
	basep->intr_mode = (basep->intr_mode & ~0x00e0) | 0x0040;
#ifdef	PORT0
	vector = (basep->intr_mode & 0x00e0) | 0x0d;	/* SCC1 Intr. vec */
#ifdef	INTERMETRIC
	*((UINT32 *) (4*vector)) = (UINT32) Intr_ReceiveFrame;
#else
	*((UINT32 *) (4*vector)) = (UINT32) _INTR_SR;	/* GNU MODE */
#endif
	basep->scc_regs[PORT].mask = HDLC_RXFR|HDLC_BUSY|HDLC_RXBD;
	basep->intr_mask |= 0x2000;	/* SCC1 interrupts */
#else	/* PORT1 */
	vector = (basep->intr_mode & 0x00e0) | 0x0a;	/* SCC2 Intr. vec */
#ifdef	INTERMETRIC
	*((UINT32 *) (4*vector)) = (UINT32) Intr_ReceiveFrame;
#else
	*((UINT32 *) (4*vector)) = (UINT32) _INTR_SR;	/* GNU MODE */
#endif
	basep->scc_regs[PORT].mask = HDLC_RXFR|HDLC_BUSY|HDLC_RXBD;
	basep->intr_mask |= 0x0400;	/* SCC2 interrupts */
#endif
#ifdef	PORT0
	basep->intr_inservice = 0x2000;
#else	/* PORT1 */
	basep->intr_inservice = 0x0400;
#endif
	old_intr = (orig_intr_priority>=3)?3:orig_intr_priority;
	_SPL(old_intr); /* turn it on */

	/* finally enable receiver/transmit hardware */
	hdlc_mp->tx_enable = 1;
	hdlc_mp->rx_enable = 1;
	init_complete = TRUE;
}

/***************************************************************************/
/*  Procedure: SetAddress
	informs all necessary structes about our current address 
*/
/***************************************************************************/
SetAddress(addr)
octet addr;
{
	UINT16 *udatap;

	hdlcp->r_addr1 = addr;			/* tell 302 recv CP process */

	udatap = (UINT16 *) &(basep->udata[0]);
	*udatap++ = addr;		/* addr 0 => ENQ */
	udatap++;
	*udatap++ = addr;		/* addr 1 => ACK */
	udatap++;
	*udatap++ = addr;		/* addr 2 => RTS */
	udatap++;
	*udatap++ = addr;		/* addr 3 => CTS */
	udatap++;
	*udatap++ = addr;		/* addr 4 => HEAD */
}

/***************************************************************************/
/*  Procedure: CarrierSense
		provides a true/false indication of CD active since last
		RstCarrierSense() call.
*/
/***************************************************************************/
CarrierSense()
{
	if( basep->scc_regs[PORT].status & SCCS_CDB)	/* if no instant carrier */
		if(  !(basep->scc_regs[PORT].event & HDLC_CD) ){ /* if no change*/
			return(0);
		}
	return(1);
}

/***************************************************************************/
/*  Procedure: la_ei
		enable 302 scc1 interrupts
*/
/***************************************************************************/
la_ei()
{
	_SPL(old_intr);
}
/***************************************************************************/
/*  Procedure: la_di
		disable 302 scc1 interrupts
*/
/***************************************************************************/
la_di()
{
	old_intr = _GPL();
	_SPL(7);		/* turn off all interrupts */
}

/***************************************************************************/
/*  Procedure: STimer
	set timer to timeout at specified val (uS)
	use GTimer to determine if timeout occured
*/
/***************************************************************************/
STimer(val)
UINT16 val;
{
	basep->t0_mode = 0xc;			/* reset timer 0 */
	basep->t0_ref = val;			/* set reference value */
	basep->t0_status = 0x3;			/* clear any events */

	basep->t0_mode = 0xd;			/* start timer */
}


/***************************************************************************/
/*  Procedure: GTimer
	determine if timer has timed out
*/
/***************************************************************************/
GTimer()
{
	if(basep->t0_status & 0x2)		/* reference event occurred? */
		return(TRUE);
	 else
		return(FALSE);
}


/******************************************************************
 *                                                                *
 *  This file is in four parts:                                   *
 *      1)  Global Variable Declarations                          *
 *      2)  Receive Routines                                      *
 *      3)  Hardware Initialization/Interface (low-level)         *
 *      4)  Transmit Routines                                     *
 *                                                                *
 *                                                                *
 *                                                                *
 *  ************************************************************  *
 *     Part 4:  Transmit Routines                                 *
 *                                                                *
 ******************************************************************/

/***************************************************************************/
/*  Procedure: la_dinit				( USER INTERFACE ROUTINE )
	driver layer
	called one time at program startup to initialize global variables
*/
/***************************************************************************/
la_dinit()
{

/* initialize any and all global variables
	=> this allows for a "warm" start */

	/* LLAP global variables */
	MyAddress=0;		/* set during InitializeLAP */
	Backoff=0;		/* current backoff range */
	fAdrValid=0;	/* MyAddress has been validated */
	fAdrInUse=0;	/* Another node has same MyAddress */
	fCTSexpected=0;		/* RTS has been sent, CTS is valid */

#ifdef	STATISTIC
	deferCount=collsnCount=0;	/* optional, for statistics only */
#endif

	deferHistory=0;
	collsnHistory=0;

	/* other global variables */
	la_recv_init();		/* init the receive variables */

/* SECOND INITIALIZE THE HARDWARE
	=> WARNING: THIS ONLY INITS THE HARDWARE, NOT THE PROTOCOL.
	I.E. a node address is NOT chosen and verified
*/

	/* initialize hardware */
	mc302_init();
}

/***************************************************************************/
/*  Procedure: InitializeLLAP			( USER INTERFACE ROUTINE )
	Called to reset LLAP's global variables
	calls AcquireAddress to initilze MyAddress
*/
/***************************************************************************/
InitializeLLAP(hint,server)
octet hint;
Boolean server;
{
	if(sdebug>2)
		printf("InitializeLLAP: hint=x%x, server=%d\n",hint,server);

	Backoff = 0;
	/* initialize history data for Backoff Calculations */
	deferHistory = 0;
	collsnHistory = 0;

#ifdef	STATISTIC
	deferCount = 0;
	collsnCount = 0;
#endif

	return(AcquireAddress(hint,server));
}


/***************************************************************************/
/*  Procedure: AcquireAddress			( DRIVER INTERNAL ROUTINE )
	Acquires unique dynamic node ID.
	creates and sends a lapENQ packet.  When no node responds after 
		repeated attempts, the current value of MyAddress is ASSUMED
		to be safe for use by this node;
	fAdrValid set true if successful
	If fAdrInUse ever becomes true after a call to AcquireAddress, then
		another node is using the same address has been detected.
*/
/***************************************************************************/
AcquireAddress(hint,server)
octet hint;
Boolean server;
{
	int maxTrys,trys;
	aFrame ENQframe;
	void *pp;
	int tmp;

	if(hint>0)
		MyAddress = hint;
	 else if(server)
		MyAddress = RANDOM(127) + 128;
	 else
		MyAddress = RANDOM(127) + 1;
	SetAddress(MyAddress);

	fAdrValid = FALSE;
	if(server)
		maxTrys = wksTries * 6; /* servers try 6 times as much as workstations*/
	  else
		maxTrys = wksTries;
	trys = 0;
	fAdrInUse = 0;

	while(trys<maxTrys) {
if(sdebug>2)
printf("trying to set network address: trys=%d maxTrys=%d\n",trys,maxTrys);
		pp = (void *) (ENQframe.structured.dataField);
		tmp = TransmitPacket(MyAddress,lapENQ,(aDataField *)pp,0);
		if( tmp==transmitOK || fAdrInUse) {
if(sdebug>1)
printf("Network Address = x%x is already in use\n",MyAddress);
			if(server)
				MyAddress = RANDOM(127) + 128;
			  else
				MyAddress = RANDOM(127) + 1;
			SetAddress(MyAddress);
			fAdrInUse = 0;
			trys = 0;
		} else
		/* ADDITION TO APPLE CODE */
		if(tmp==notInitialize || tmp==networkError) {
if(sdebug>0)
printf("ERROR: Cannot initialize driver, error=%d/%s\n",tmp,lapresult[tmp]);
			return(tmp);
		/* END OF ADDITION */
		} else
			trys++;
	}
if(sdebug>0)
printf("setting Address Valid Flag;  Net Address is x%x\n",MyAddress);
	fAdrValid = 1;
	return(OKinitialize);
}


/***************************************************************************/
/*  Procedure: TransmitPacket			( USER INTERFACE ROUTINE )
	Called by LLAP client to a send a Data Packet
	after encapsulating the dataParam, TransmitLinkMgmt
		to actually send the packet
*/
/***************************************************************************/
TransmitPacket(dstParam,typeParam,dataParam,dataLength)
anAddress dstParam;
aLAPtype typeParam;
aDataField *dataParam;
int dataLength;
{
	if(sdebug>2)
		printf("In TransmitPacket: dst=x%x type=x%x length=%d\n",
					dstParam,typeParam,dataLength);

	if(!init_complete) {
		if(sdebug>1)
			printf("la driver hardware not initialized!!\n");
		return(notInitialize);;
	}

	/** ADDED TO APPLE CODE **/
	if(fAdrInUse) {
		if(sdebug>2)
			printf("ERR: Address In Use is set\n");
		return(dupAddress);
	}
	/** END OF ADDITION */

	/* configure header */
	basep->pram[PORT].scc.tbd[XBD_HEAD].buf[0] = dstParam;
	basep->pram[PORT].scc.tbd[XBD_HEAD].buf[2] = typeParam;
	
	/* point to user data */
	basep->pram[PORT].scc.tbd[XBD_DATA].buf = (UCHAR *)dataParam;
	basep->pram[PORT].scc.tbd[XBD_DATA].length = dataLength;
	
	return(TransmitLinkMgmt());
}



/***************************************************************************/
/*  Procedure: TransmitLinkMgmt			( DRIVER INTERNAL ROUTINE )
	Implements the actual Carrier Sense Multiple Access with Collision Avoidance
	LLAP attempts to minimize collisions by requiring transmitters to wait for
		the duration of the interdialog gap (IDG) plus a random period of time
		before sending their request-to-send (RTS) packets.  Any transmitter
		that detects that another transmission is in progress while it is
		waiting must defer.
*/
/***************************************************************************/
TransmitLinkMgmt()
{
	int LclBackoff,BackOffFlag;
	Boolean fBroadcast,fENQ;
	int deferTries,collsnTries;		/* ???? */

	fBroadcast = (basep->pram[PORT].scc.tbd[XBD_HEAD].buf[0] == BROADCAST) 
				? TRUE : FALSE ;
	fENQ = (basep->pram[PORT].scc.tbd[XBD_HEAD].buf[2] == lapENQ) 
				? TRUE : FALSE;

	/* Adjust Backoff, based upon recent history */
	if( ones(collsnHistory) > 2 ) {
		/* Increase Backoff it we've seen a lot of collisions */
		Backoff = MIN(MAX(Backoff<<1,2),16);
		collsnHistory = 0x00;
	} else
	if( ones(deferHistory) <2 ) {
		/* Decrease Backoff if we haven't had to defer very much */
		Backoff >>= 1;
		deferHistory = 0xff;
	}

	if(sdebug>2)
		printf("new Backoff=x%x defHistory=x%x\n",Backoff,deferHistory);
	/* shift history data */
	collsnHistory <<= 1;
	deferHistory <<=1;

	/* initialize main loop */
	deferTries = collsnTries = 0;
	LclBackoff = Backoff ;

	/* begin main loop */
	while(1) {
		/* wait for minimum Inter-Dialog Gap time */
		BackOffFlag=1;
		while(BackOffFlag!=0) {
			RstCarrierSense();
			if( CarrierSense() ) {	/* but first wait for end of current frame*/
				LclBackoff = MAX(LclBackoff,2);
				deferHistory |= 0x01 ;
				/* wait 1.5 maxFrameSize times */
				STimer((UINT16)( (1.5 * maxFrameSize *8)*bitTime ));
				while(CarrierSense()) {
					if(GTimer()) { /* network is locked up*/
						/* in apple this is a 8530 problem coniditon
						in 68302 only network problems are known
						to trip this condition */
						return(networkError);
					}
					RstCarrierSense();
				}
			}
			/* wait for minimum IDG after packet (or idle line) */
			STimer(minIDGtime);
			while( !CarrierSense() ) {
				if(GTimer()) {
					/* don't have carrier and haven't for IDG */
					BackOffFlag = 0;
					break;
				}
			}
		}
		/* wait our additional backoff time, deferring to others */
		/* (LclBackoff-1) is the upper bound of the random # to pick */
		if(sdebug>1)
			printf("LclBackoff=%d\n",LclBackoff);
		if(LclBackoff>0) {
			STimer( (UINT16) (RANDOM(LclBackoff)*IDGslottime) );
			while( !CarrierSense() )
				if(GTimer())
					break;
		}
		if( CarrierSense()) {
			if(sdebug>2)
				printf("defering\n");
			/* defer */
#ifdef	STATISTIC
			deferCount++;
#endif
			LclBackoff = MAX(LclBackoff,2);
			deferHistory |= 0x01;
			if(deferTries <maxDefers)
				deferTries++;
			  else
				return(excessDefers);
		} else	{
			RX_Frame_Status = 0; /* semaphore back from recv intr*/
			if(fENQ) {
				basep->pram[PORT].scc.tbd[XBD_ENQ].buf[0] =
					basep->pram[PORT].scc.tbd[XBD_HEAD].buf[0];
				TransmitFrame(XBD_ENQ);
			}  else	{
				basep->pram[PORT].scc.tbd[XBD_RTS].buf[0] =
					basep->pram[PORT].scc.tbd[XBD_HEAD].buf[0];
				TransmitFrame(XBD_RTS);
			}

			/* wait for response to our ENQ or RTS */
			/* to match apple driver, this is in two parts: */
			/* part 1: minIDGtime for carrier */
			fCTSexpected = TRUE; /* semaphore to recv intr */
			STimer(minIDGtime);
			RstCarrierSense();
			while( !CarrierSense() )
				if(GTimer())
					break;
			/* part 2: expect packet complete in 
				maxIFG + CTS packettime (not open flag or
					dest time) */

			if(CarrierSense()) {
				/* original apple code had just ifg before a
				character level interrupt, we get a packet
				interrupt; so we wait ifg + remainder of frame
				(which is SrcAddr,type,2bytesCRC,1close flag) */
				/* don't forget possible bit-stuffing on all but flag */
				/* 201=remainder of frame */
				/* we used to to have this at 190 */
				STimer(maxIFGtime + 201);
				while(RX_Frame_Status==0) { /* wait for receive frame*/
					if(GTimer())
						break;
				}
			}
			/* RX_Frame_Status = 0 if no framed received,
				else equals frame received type */
			fCTSexpected = FALSE;

#ifdef TIMING
basep->pa_data=DATAMARK;
basep->pa_data=DATADEF;
#endif

			if(fAdrInUse) {
				if(sdebug>2)
					printf("fAdrInUse set in transmit due to rcv pkt\n");
				return(dupAddress);
			} else	{
				switch(RX_Frame_Status) {
				case 0:			/* timeout */
					if(fBroadcast) {
						TransmitFrame(XBD_HEAD);
						return(transmitOK);
					}
					break;
				case lapDATA:
					/* DATA packet was successfully sent
					by the receive interrupt handler */

					return(transmitOK);
				}
			}

			/* Assume collision if we don't receive the expected CTS */
			if(sdebug>2)
				printf("have a collision\n");
#ifdef	STATISTIC
			collsnCount++;
#endif
			collsnHistory |= 0x01;
			collsnTries++;
			if( collsnTries < maxCollsns) {
				LclBackoff = MIN(MAX(LclBackoff<<1,2),16);
			} else {
				if(sdebug>2)
					printf("have excess collisions\n");
				return(excessCollsns);
			}
			if(sdebug>2)
				printf("collsnHistory=x%x tries=%d\n",
						collsnHistory,collsnTries);
		}
	}
}

/***************************************************************************/
/*  Procedure: TransmitFrame
	Puts the data on the actual link
*/
/***************************************************************************/
TransmitFrame(type)
int type;			/* id of frame type to transmit */
{
	struct hdlc_mode_reg *hdlc_mp;
	int oldtstate;

	if(sdebug>4)
		printf("In TransmitFrame\n");

	/* disable Receiver */
	hdlc_mp = (struct hdlc_mode_reg *) &(basep->scc_regs[PORT].mode);
	hdlc_mp->rx_enable = 0;
	la_di();
	while(basep->cr & 1);		/* wait for command idle */
	basep->scc_regs[PORT].event = HDLC_RXFR|HDLC_BUSY|HDLC_RXBD;
	
	oldtstate = basep->pram[PORT].scc.pscc.h.tstate;
	if(type==XBD_HEAD)	{ 	/* data packet */
		if(basep->pram[PORT].scc.tbd[XBD_DATA].length==0) {
			basep->pram[PORT].scc.tbd[XBD_HEAD].status |= (T_W|T_L);
		} else	{
			basep->pram[PORT].scc.tbd[XBD_DATA].status |= T_R;
		}
	}
	basep->pram[PORT].scc.tbd[type].status |= T_R; /* send frame */
	basep->pram[PORT].scc.pscc.h.tbuf = 0x40 + (type<<3);

#ifdef	TIMING
basep->pa_data=DATAMARK;
#endif

	hdlc_mp->flg = 1;
	while(oldtstate==basep->pram[PORT].scc.pscc.h.tstate) ;
	hdlc_mp->flg = 0;
	basep->cr = ENTER_HUNT_MODE | (PORT<<1) | 1;

	/* wait for transmission */
	if(type==XBD_HEAD)	{ 	/* data packet */
		if(basep->pram[PORT].scc.tbd[XBD_DATA].length==0) {
			while(basep->pram[PORT].scc.tbd[XBD_HEAD].status & T_R);
		} else
			while(basep->pram[PORT].scc.tbd[XBD_DATA].status & T_R);
	} else
		while(basep->pram[PORT].scc.tbd[type].status & T_R);

	/* 302 guanrantees frame completely transmitted before clearing
		buffer descriptor.  This concerns the data field only.
		closing flag and the idle sequence may still be being sent.*/

	/* enable Receiver */
	while(basep->cr & 1); /* wait for hunt completion */
	hdlc_mp->rx_enable = 1;

	/* wait for frame trailer IDLE-seq */
		/* incls la 1.5 squelch timer and la internal delays */
	for(STimer(88);!GTimer(););
	basep->pram[PORT].scc.tbd[XBD_HEAD].status &= ~(T_W|T_L);
	la_ei();

#ifdef	TIMING
basep->pa_data=DATAMARK;
basep->pa_data=DATADEF;
#endif

}





