/******************************************************************
 * Copyright (c) 1990  MOTOROLA, Inc.                             *
 * NOTICE: The information contained in this file is proprietary  *
 * to MOTOROLA, INC. and is being made available  to MOTOROLA'S   *
 * customers under strict license agreements.              *
 * Use or disclosure of this information is permissible only      *
 * under the terms of the existing license agreement.      *
 ******************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "system.h"				/* variable type defines */
#include "imp.h"
#include "hdlc.h"
#include "la.h"

Boolean echo_enable[256];


/***************************************************************************/
/*  Procedure: la_echoinit
	initializes the structures associated with the la_echo routine
	below
*/
/***************************************************************************/
la_echoinit()
{
	int i;

	for(i=0;i<256;i++)
		echo_enable[i]=FALSE;
}

/***************************************************************************/
/*  Procedure: la_echo				USER INTERFACE ROUTINE
	controls the echo back onto the network of received packets
	packets types can be selected or deselected from being echoed.
	once selected any packet receive by the software is transferred
	(as a background process) to transmit packet with the source
	and destination addresses reversed (loopback).

	NOTE:  The echo capability is provided as "application" layer
	that rides on top of the link layer.  This sample driver software
	does NOT provide a datagram delivery layer, as such echo competes
	with the Receive command for the receive packets.

	Echo will function as expected when other commands are not being
	executed.  Predictible, but perhaps surprising, results will occur
	if echo is activated while other command are utilized.
*/
/***************************************************************************/
la_echo(argc,argv)
int argc;
char *argv[];
{
	register i;
	int response;
	Boolean flag,chgflag;
	Boolean echoen[256];		/* local copy for editing */
	int c;
	extern int optind;
	extern char *optarg;


	for(i=0;i<256;i++)
		echoen[i] = echo_enable[i];
	chgflag = FALSE;

	while((c=getopt(argc,argv,"a:d:csA:D:CS")) != EOF)
		switch(c) {
		case 'a':
		case 'A':
			echoen[aton(optarg)%256] = TRUE;
			chgflag = TRUE;
			break;
		case 'd':
		case 'D':
			echoen[aton(optarg)%256] = FALSE;
			chgflag = TRUE;
			break;
		case 'c':
		case 'C':
			for(i=0;i<256;i++)
				echoen[i] = FALSE;
			chgflag = TRUE;
			break;
		case 's':
		case 'S':
			for(i=0;i<256;i++)
				echoen[i] = TRUE;
			chgflag = TRUE;
			break;
		case '?':
			help(0,(char **)0,"echo");
			return;
		}
	if(chgflag) {
		for(i=0;i<256;i++)
			echo_enable[i] = echoen[i];
		la_echop();
		return;
	}

	la_echop();
	putchar('\n');
	while(1) {
		printf("Do you wish to add/delete a packet type [n]: " );
		if( !getyesno()) {		/* all done */
			if(chgflag==TRUE) {
				for(i=0;i<256;i++)
					echo_enable[i] = echoen[i];
				la_echop();
			}
			return;
		}
		for(flag=TRUE;flag==TRUE;) {
			printf("Enter a/add, d/delete, c/clear all, s/set all, x/exit: ");
			response = getchar();
			while(getchar()!='\n');
			flag=FALSE;		/* assume good */
			switch(tolower(response)) {
			case 'a':
				printf("Enter packet type to add [0]: ");
				response = readnum();
				echoen[response%256] = TRUE;
				chgflag = TRUE;
				break;
			case 'd':
				printf("Enter packet type to delete [0]: ");
				response = readnum();
				echoen[response%256] = FALSE;
				chgflag = TRUE;
				break;
			case 'c':
				for(i=0;i<256;i++)
					echoen[i] = FALSE;
				chgflag = TRUE;
				break;
			case 's':
				for(i=0;i<256;i++)
					echoen[i] = TRUE;
				chgflag = TRUE;
				break;
			case 'x':
				return;
			default:
				flag=TRUE;	/* oops! */
			}
		}
	}
}

la_echop()
{
	int i;
	int noflag,yesflag;

	printf("The following packet types are currently slated for loopback:\n");
	yesflag = noflag = 1;
	for(i=0;i<256;i++) {		/* determine if none or all */
		if(echo_enable[i] == TRUE)
			noflag=0;
		if(echo_enable[i] == FALSE)
			yesflag=0;
	}
	if(noflag==TRUE) {
		printf("\tNONE: no packets types are currently slated\n");
		return;
	}
	if(yesflag==TRUE) {
		printf("\tALL: all packets types are currently slated\n");
		return;
	}
	for(i=0;i<256;i++)
		if(echo_enable[i] == TRUE) {
			printf(" x%x",i);
		}
	putchar('\n');
}

/***************************************************************************/
/*  Procedure: la_echobkgrd
	actual background routine that scans the receive buffer link list
	looking for dealing with packets to be echoed
*/
/***************************************************************************/
la_echobkgrd()
{
	extern struct rbuf rbuffer[MAXRBUF];	/* allocate receive buffers */
	extern struct rbuf freelist;		/* header for free list */
	extern struct rbuf rhdr;		/* header for receive list */
	extern struct rbuf rtail;		/* tail pointer for receive list */
	struct rbuf *p,*lp;
	void *pp;
	int result;

	for(lp=p=rhdr.next;p!=(struct rbuf *)0;) {	   /* scan list */
		if(echo_enable[p->rframe.structured.lapType] == TRUE) {
			/* have a packet to be echoed */
			/* remove from list */
			la_di();		/* disable interrupts */
			if(p==rhdr.next) {	/* 1st item on list */
				rhdr.next = p->next;
				if(rtail.next==p)  /* 1st and last */
					rtail.next = (struct rbuf *)0;
			} else	{
				lp->next = p->next;
				if(rtail.next==p)  /* last */
					rtail.next=lp;
			}
			la_ei();		/* re-enable interrupts */

			/* take out addresses,type,crc from length */
			/* should be in interrupt, but here reduces interrupt time */
			p->length = ((INT16)(p->length)) - 5;
			if( ((INT16)(p->length)) <0)
				p->length = 0;

			/* transmit packet */
			pp = (void *) p->rframe.structured.dataField;
			if( (result=TransmitPacket(
					p->rframe.structured.srcAddr,
					p->rframe.structured.lapType,
		   			(aDataField *)pp,
					(int) (p->length)))
						 != transmitOK) {
				printf("\n loopback error, result=x%x\n",result);
			}
			FreePacket(p);   /* return packet buffer */
			if(p==lp)	/* 1st entry removed */
				lp=p=rhdr.next;
			  else
				p = lp->next;
			break;		/* do only one packet per call */
		} else	{
			lp = p;
			p = p->next;
		}
	}
}


