/******************************************************************
 * Copyright (c) 1990  MOTOROLA, Inc.                             *
 * NOTICE: The information contained in this file is proprietary  *
 * to MOTOROLA, INC. and is being made available  to MOTOROLA'S   *
 * customers under strict license agreements.              *
 * Use or disclosure of this information is permissible only      *
 * under the terms of the existing license agreement.      *
 ******************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "system.h"				/* variable type defines */
#include "imp.h"
#include "hdlc.h"
#include "la.h"



/***************************************************************************/
/*  Procedure: la_etest				USER INTERFACE ROUTINE
	performs a test of the net (and driver software) over the
	network by repeatedly sending packets over the net and having
	a remote node echo them back (via la_echo).

	NOTE:  The etest capability is provided as "application" layer
	that rides on top of the link layer.  This sample driver software
	does NOT provide a datagram delivery layer, as such etest competes
	with the Receive command for the receive packets.

	Etest will function as expected when other commands are not being
	executed.  Predictible, but perhaps surprising, results will occur
	if etest is activated while other command are utilized.
*/
/***************************************************************************/
la_etest(argc,argv)
int argc;
char *argv[];
{
	int dest,rpkt,pkttype;
	int rpktdata;
	octet xbuf[maxDataSize];
	int i,j,pktlength;
	void *pp;
	int repeat;
	int terror,rerror,echoerror,conerr;
	int eflag;
	struct rbuf *p;

	if( argc != 1) {
		help(0,(char **)0,"etest");
		return;
	}
	if(fAdrValid==0) {
		printf("Driver has not been initialized\n");
		return;
	}

	printf("Perform link test via external remote loopback\n\n");
	terror=rerror=echoerror=conerr=0;
	eflag=0;		/* has an error occurred ? */

	/* give user opportunity to abort command */
	printf("Do you wish to abort this operation [n]: ");
	if(getyesno())
		return;

	/* remote node information */
	for(dest= -1;dest<=0 || dest>=255;) {
		/* do not allow null or broadcast addresses */
		printf("Enter Remote node Address [1-255]: ");
		dest = readnum();
	}

	/* packet type information */
	printf("Do you wish to send a randomly selected packet type [n]: ");
	if(getyesno()) {		/* yes, set random */
		rpkt = 1;
	} else {			/* no, use a fixed packet type */
		rpkt = 0;
		for(pkttype= -1;pkttype<1 || pkttype>=0x80;) {
			printf("Enter packet type to use [1-127]: ");
			pkttype = readnum();
		}
	}

	/* data field information */
	printf("Do you wish to send a randomly selected data field [n]: ");
	if(getyesno()) {		/* yes, set random */
		for(rpktdata= -1;rpktdata<0 || rpktdata>maxDataSize;) {
			printf("Enter length of random data field [0-%d] (0/random length) [0]: ",
					maxDataSize);
			rpktdata = readnum();
		}
	} else	{			/* no, use a entered frame */
		rpktdata = -1;
		printf("Enter the data field to be sent (CR to end): ");
		gets(xbuf);
		pktlength = strlen(xbuf);
	}

	/* # of tries information */
	for(repeat= -1;repeat<0;) {
		printf("Enter number of times to perform test (0 for continuous) [0]: ");
		repeat = readnum();
	}

	/* done with information gathering */
	/* prompt user to configure remote node */
	printf("\nIn order for this test properly function,\n");
	printf("\tthe remote node must be correctly configured.\n\n");
	printf("Please configure node x%x to echo-back ",dest);
	if(rpkt!=0)
		printf("all packet types.\n");
	  else
		printf("packet type: x%x\n",pkttype);
	printf("Enter <cr> to continue test, y to abort test [<cr>]: ");
	if(getyesno())
		return;

	/* now actually perform the test */
	pp = (void *)xbuf;

	for(i=1;i<=repeat || repeat==0;i++) {
		if(cbreak()) {
			printf("\n User has aborted test!\n");
			break;		/* premature exit, user struck a key */
		}
		/* make up the random stuff */
		if(rpkt!=0)
			pkttype = rand()%0x80;
		if(rpktdata==0)
			pktlength = rand()%(maxDataSize+1);
		if(rpktdata>=0) 
			for(j=0;j<pktlength;j++)
				xbuf[j] = rand()&0xff;

		if( TransmitPacket((anAddress)dest,(aLAPtype)pkttype,
			(aDataField *)pp,pktlength) != transmitOK) {
			eflag=1;
			terror++;
			goto nexttry;
		}

		/* wait for echo back */
		basep->t1_mode = 0xff0c;	/* 256uS/tick */
		basep->t1_ref = 0x8000;		/* 16.67/2 sec timeout */
		basep->t1_status = 0x3;		/* clear any events */
		basep->t1_mode = 0xff0d;		/* start timer */
		while(RPacketAvail() == 0) {		/* wait for receive packet */
			if( (basep->t1_status & 0x2)!=0 ) {	/* timeout */
				eflag=1;
				echoerror++;
				goto nexttry;
			}
		}

		/* get received packeted */
		p = RPacket();
		/* compare received packet */
		if(p->length != pktlength) {
			eflag=1;
			rerror++;
			goto nexttry1;
		}
		for(j=0;j<pktlength;j++)
			if(xbuf[j] != p->rframe.structured.dataField[j]) {
				eflag=1;
				rerror++;
				goto nexttry1;
			}
		/* here if good echo back */

		/* though, we don't do anything special on good (only bad) */
nexttry1:
		FreePacket(p);
nexttry:
		/* display any stats */
		if( i<100 || (i%100)==0 || eflag!=0)
			printf("[%5d] transmit errs=%3d  echo error=%3d receive error=%3d    \r",
				i,terror,echoerror,rerror);
		/* check too many errors */
		if(eflag)
			conerr++;
		  else
			conerr=0;
		eflag=0;
		if(conerr>50) {
			printf("\n Too many consecutive errors, test exited\n");
			return;
	}
	}
	putchar('\n');
}


