/******************************************************************
 * Copyright (c) 1990  MOTOROLA, Inc.                             *
 * NOTICE: The information contained in this file is proprietary  *
 * to MOTOROLA, INC. and is being made available  to MOTOROLA'S   *
 * customers under strict license agreements.              *
 * Use or disclosure of this information is permissible only      *
 * under the terms of the existing license agreement.      *
 ******************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "system.h"				/* variable type defines */
#include "imp.h"
#include "hdlc.h"
#include "la.h"



#define	LAPTYPE		1	/* use short DDP format */
#define	AEPSOCKET	4	/* aep socket number */
#define	DDPTYPE		4	/* ddp type field */
#define	maxAEPSize	585	/* max # of data bytes in AEP packet */

#define	HEADER_SIZE	6	/* combined size of DDP and AEP headers */

#define	AEP_RQ		1	/* request definition (AEP header) */
#define	AEP_RPLY	2	/* reply definition (AEP header) */

/***************************************************************************/
/*  Procedure: la_mtest				USER INTERFACE ROUTINE
	performs a test of the net (and driver software) over the
	network by repeatedly sending packets over the net and having
	a remote node (running the appletalk protocol stack) echo them
	back.

	mtest (MAC-TEST) is identical to etest except that the
	appropriate AEP (Apple Echo Protocol) DDG (DataGram) packet 
	is generated that causes a remote mac to echo back the transmitted
	packet.  this test insures that the driver is indeed capable
	of communicating with a remote apple protocol stack.

	NOTE:  The mtest capability is provided as "application" layer
	that rides on top of the link layer.  This sample driver software
	does NOT provide a datagram delivery layer, as such mtest competes
	with the Receive command for the receive packets.

	Etest will function as expected when other commands are not being
	executed.  Predictible, but perhaps surprising, results will occur
	if mtest is activated while other command are utilized.
*/
/***************************************************************************/
la_mtest(argc,argv)
int argc;
char *argv[];
{
	int dest,rpkt,pkttype;
	int rpktdata;
	octet xbuf[maxFrameSize];
	int i,j,k;
	UINT16 pktlength;
	void *pp;
	int repeat;
	int terror,rerror,echoerror,conerr;
	int eflag;
	struct rbuf *p;

	if( argc != 1) {
		help(0,(char **)0,"mtest");
		return;
	}
	if(fAdrValid==0) {
		printf("Driver has not been initialized\n");
		return;
	}

	printf("Perform APPLETALK link test via external remote loopback\n\n");
	printf("  Note: This test requires another node running the apple stack\n");
	terror=rerror=echoerror=conerr=0;
	eflag=0;		/* has an error occurred ? */

	/* give user opportunity to abort command */
	printf("Do you wish to abort this operation [n]: ");
	if(getyesno())
		return;

	/* remote node information */
	for(dest= -1;dest<=0 || dest>=255;) {
		/* do not allow null or broadcast addresses */
		printf("Enter Remote Macintosh node Address [1-255]: ");
		dest = readnum();
	}

	/* packet type information */
	rpkt =0;
	pkttype = 1;

	/* data field information */
	printf("Do you wish to send a randomly selected data field [n]: ");
	if(getyesno()) {		/* yes, set random */
		for(rpktdata= -1;rpktdata<0 || rpktdata>maxAEPSize;) {
			printf("Enter length of random data field [0-%d] (0/random length) [0]: ",
					maxAEPSize);
			rpktdata = readnum();
		}
	} else	{			/* no, use a entered frame */
		rpktdata = -1;
		printf("Enter the data field to be sent (CR to end): ");
		gets( &(xbuf[HEADER_SIZE]) );
		pktlength = strlen( &(xbuf[HEADER_SIZE]) );
	}

	/* # of tries information */
	for(repeat= -1;repeat<0;) {
		printf("Enter number of times to perform test (0 for continuous) [0]: ");
		repeat = readnum();
	}

	/* done with information gathering */
	printf("\nEnter <cr> to perform test, y to abort test [<cr>]: ");
	if(getyesno())
		return;

	/* fill in fixed stuff */
	/* remember xbuf contains DDP packet, not a LAP or AEP frame */
	xbuf[2] = AEPSOCKET;
	xbuf[3] = 7;		/* doesn't matter what source socket is */
	xbuf[4] = DDPTYPE;
	xbuf[5] = AEP_RQ;

	/* now actually perform the test */
	pp = (void *)xbuf;

	for(i=1;i<=repeat || repeat==0;i++) {
		if(cbreak()) {
			printf("\n User has aborted test!\n");
			break;		/* premature exit, user struck a key */
		}
		/* make up the random stuff */
		if(rpkt!=0)
			pkttype = rand()%0x80;
		if(rpktdata==0)
			pktlength = rand()%(maxAEPSize+1);
		if(rpktdata>=0) 
			for(j=0;j<pktlength;j++)
				xbuf[j+HEADER_SIZE] = rand()&0xff;

		*((UINT16 *)pp) = pktlength + HEADER_SIZE;

		if( TransmitPacket((anAddress)dest,(aLAPtype)pkttype,
			(aDataField *)pp,((int)pktlength+HEADER_SIZE))!=transmitOK) {
			eflag=1;
			terror++;
			goto nexttry;
		}

		/* wait for echo back */
		/* use the second timer */
		basep->t1_mode = 0xff0c;	/* 256uS/tick */
		basep->t1_ref = 0x8000;		/* 16.67/2 sec timeout */
		basep->t1_status = 0x3;		/* clear any events */
		basep->t1_mode = 0xff0d;		/* start timer */
		while(RPacketAvail() == 0) {		/* wait for receive packet */
			if( (basep->t1_status & 0x2)!=0 ) {	/* timeout */
				eflag=1;
				echoerror++;
				goto nexttry;
			}
		}

		/* get received packeted */
		p = RPacket();

		/* compare received packet */
		if(p->length != (pktlength+HEADER_SIZE) ) {
			eflag=1;
			rerror++;
			goto nexttry1;
		}
		for(j=0;j< (pktlength+HEADER_SIZE) ;j++) {
			/* check AEP Type specially */
			if(j==5)
				if(p->rframe.structured.dataField[j] != AEP_RPLY) {
					goto err;
				} else
					continue;
			/* swap comparision of src&dest sockets */
			k=j;
			if(j==2)
				k=3;
			if(j==3)
				k=2;
			if(xbuf[k] != p->rframe.structured.dataField[j]) {
err:
				eflag=1;
				rerror++;
				goto nexttry1;
			}
		}
		/* here if good echo back */

		/* though, we don't do anything special on good (only bad) */
nexttry1:
		FreePacket(p);
nexttry:
		/* display any stats */
		if( i<100 || (i%100)==0 || eflag!=0)
			printf("[%5d] transmit errs=%3d  echo error=%3d receive error=%3d    \r",
				i,terror,echoerror,rerror);
		/* check too many errors */
		if(eflag)
			conerr++;
		  else
			conerr=0;
		eflag=0;
		if(conerr>50) {
			printf("\n Too many consecutive errors, test exited\n");
			return;
		}
	}
	putchar('\n');
}



